from . import util

def radify_cat(cat, col):
 op = "IF ISEMPTY(" + str(col) + ") THEN " + str(cat["OTHER"]) + " ELSE "
 for u in cat["uniques"]:
  if type(u)==str:
   op += ("IF " + col + '="' + u + '" THEN ' +  str(cat["uniques"][u]) + " ELSE ")
  else:
   op += ("IF " + col + '=' + str(util.denump(u)) + ' THEN ' +  str(cat["uniques"][u]) + " ELSE ")
 op += (str(cat["OTHER"]))
 return op

def radify_cont(cont, col, defaultValue=1):
 op = "IF ISEMPTY(" + str(col) + ") THEN " + str(defaultValue) + " ELSE IF ISNAN(" + str(col) + ") THEN " + str(defaultValue) + " ELSE "
 op += ("IF " + col + "<=" + str(cont[0][0]) + " THEN " + str(cont[0][1]) + " ELSE ")
 for i in range(len(cont)-1):
  p1 = str(cont[i][0])
  p2 = str(cont[i+1][0])
  y1 = str(cont[i][1])
  y2 = str(cont[i+1][1])
  op += ("IF " + col +  "<=" + p2 + " THEN " + "(("+col + "-" + p1+")*" + y2 + "+(" + p2 + "-" + col + ")*" + y1 + ")/(" + p2 + "-" + p1 + ") ELSE ") #((x-p1)y2 + (p2-x)y1) / (p2 - p1)
 op += ("IF " + col + ">=" + str(cont[-1][0]) + " THEN " + str(cont[-1][1]) + " ELSE " +str(defaultValue))
 return op

def radify_catcat(model, cols):
 col1, col2 = cols.split(" X ")
 catcat=model["catcats"][cols]
 
 op = "IF ISEMPTY(" + str(col1) + ") THEN (" + radify_cat(catcat["OTHER"], col2) + ") ELSE IF ISNAN(" + str(col1) + ") THEN " + radify_cat(catcat["OTHER"], col2) + " ELSE "
 for u in catcat["uniques"]:
  op += (" IF " + str(col1) + "= " + str(u) + " THEN (")
  op += radify_cat(catcat["uniques"][u], col2)
  op += (") ELSE ")
 op+=("("+radify_cat(catcat["OTHER"], col2)+")")
 return op

def radify_catcont(model, cols, defaultValue=1):
 col1, col2 = cols.split(" X ")
 catcont=model["catconts"][cols]
 
 op = "IF ISEMPTY(" + str(col1) + ") THEN (" + radify_cont(catcont["OTHER"], col2, defaultValue) + ") ELSE IF ISNAN(" + str(col1) + ") THEN " + radify_cont(catcont["OTHER"], col2, defaultValue) + " ELSE "
 for u in catcont["uniques"]:
  op += (" IF " + str(col1) + "= " + str(u) + " THEN (")
  op += radify_cont(catcont["uniques"][u], col2, defaultValue)
  op += (") ELSE ")
 op+= ("("+radify_cont(catcont["OTHER"], col2, defaultValue)+")")
 return op

def radify_contcont(model, cols, defaultValue=1):
 
 #we are using x and y to predict z
 
 col1, col2 = cols.split(" X ")
 contcont=model["contconts"][cols]
 
 op = ("IF ISEMPTY(" + str(col1) + ") THEN " + str(defaultValue) + " ELSE IF ISNAN(" + str(col1) + ") THEN " + str(defaultValue) + " ELSE ")
 op += ("IF ISEMPTY(" + str(col2) + ") THEN " + str(defaultValue) + " ELSE IF ISNAN(" + str(col2) + ") THEN " + str(defaultValue) + " ELSE ")
 
 #CORNERS
 
 op += ("IF " + str(col1) + "<=" + str(contcont[0][0]) + " AND " + str(col2) + "<=" + str(contcont[0][1][0][0]) + " THEN " + str(contcont[0][1][0][1]) + " ELSE ")
 op += ("IF " + str(col1) + ">=" + str(contcont[-1][0]) + " AND " + str(col2) + "<=" + str(contcont[-1][1][0][0]) + " THEN " + str(contcont[-1][1][0][1]) + " ELSE ")
 op += ("IF " + str(col1) + "<=" + str(contcont[0][0]) + " AND " + str(col2) + "<=" + str(contcont[0][1][-1][0]) + " THEN " + str(contcont[0][1][-1][1]) + " ELSE ")
 op += ("IF " + str(col1) + "<=" + str(contcont[-1][0]) + " AND " + str(col2) + "<=" + str(contcont[-1][1][-1][0]) + " THEN " + str(contcont[-1][1][-1][1]) + " ELSE ")
 
 #EDGES
 
 for i in range(len(contcont)-1):
  x1 = str(contcont[i][0])
  x2 = str(contcont[i+1][0])
  z1 = str(contcont[i][1][0][1])
  z2 = str(contcont[i+1][1][0][1])
  op+=("IF " + str(col2) + "<=" + str(contcont[i][1][0][0]) + " AND " + str(col1) + ">=" + str(x1) + " AND " + str(col1) + "<=" + str(x2) + " THEN ")
  op+=("(("+col1 + "-" + x1+")*" + z2 + "+(" + x2 + "-" + col1 + ")*" + z1 + ")/(" + x2 + "-" + x1 + ") ELSE ")
 
 for i in range(len(contcont)-1):
  x1 = str(contcont[i][0])
  x2 = str(contcont[i+1][0])
  z1 = str(contcont[i][1][-1][1])
  z2 = str(contcont[i+1][1][-1][1])
  op+=("IF " + str(col2) + ">=" + str(contcont[i][1][-1][0]) + " AND " + str(col1) + ">=" + str(x1) + " AND " + str(col1) + "<=" + str(x2) + " THEN ")
  op+=("(("+col1 + "-" + x1+")*" + z2 + "+(" + x2 + "-" + col1 + ")*" + z1 + ")/(" + x2 + "-" + x1 + ") ELSE ")
 
 for i in range(len(contcont[0][1])-1):
  y1 = str(contcont[0][1][i][0])
  y2 = str(contcont[0][1][i+1][0])
  z1 = str(contcont[0][1][i][1])
  z2 = str(contcont[0][1][i+1][1])
  op+=("IF " + str(col1) + "<=" + str(contcont[0][0]) + " AND " + str(col2) + ">=" + str(y1) + " AND " + str(col2) + "<=" + str(y2) + " THEN ")
  op+=("(("+col2 + "-" + y1+")*" + z2 + "+(" + y2 + "-" + col2 + ")*" + z1 + ")/(" + y2 + "-" + y1 + ") ELSE ")
 
 for i in range(len(contcont[0][1])-1):
  y1 = str(contcont[-1][1][i][0])
  y2 = str(contcont[-1][1][i+1][0])
  z1 = str(contcont[-1][1][i][1])
  z2 = str(contcont[-1][1][i+1][1])
  op+=("IF " + str(col1) + ">=" + str(contcont[-1][0]) + " AND " + str(col2) + ">=" + str(y1) + " AND " + str(col2) + "<=" + str(y2) + " THEN ")
  op+=("(("+col2 + "-" + y1+")*" + z2 + "+(" + y2 + "-" + col2 + ")*" + z1 + ")/(" + y2 + "-" + y1 + ") ELSE ")
 
 #INTERIOR
 
 for i in range(len(contcont)-1):
  x1 = str(contcont[i][0])
  x2 = str(contcont[i+1][0])
  for j in range(len(contcont[i][1])-1):
   y1 = str(contcont[0][1][j][0])
   y2 = str(contcont[0][1][j+1][0])
   z11 = str(contcont[i][1][j][1])
   z12 = str(contcont[i][1][j+1][1])
   z21 = str(contcont[i+1][1][j][1])
   z22 = str(contcont[i+1][1][j+1][1])
   op+=("IF " + col1 + ">=" + x1 + " AND " + col1 + "<=" + x2 + " AND " + col2 + ">=" + y1 + " AND " + col2 + "<=" + y2 + " THEN ")
   op+=("(("+col1+"-"+x1+")*("+col2+"-"+y1+")*"+z22+" + ("+col1+"-"+x1+")*("+y2+"-"+col2+")*"+z21+" + ("+x2+"-"+col1+")*("+col2+"-"+y1+")*"+z12+" + ("+x2+"-"+col1+")*("+y2+"-"+col2+")*"+z11+")/(("+x2+"-"+x1+")*("+y2+"-"+y1+")) ELSE ")
 op+=str(defaultValue)
 return op

def radify_model(model, filename="rad_model.txt", logistic=False):
 lines = ["BASE_VALUE = " + str(model["BASE_VALUE"]) + ";"]
 combLine = "COMB = BASE_VALUE"
 
 fc="*"
 defaultValue=1
 if "featcomb" in model:
  if model["featcomb"]=="addl":
   fc="+"
   defaultValue=0
 
 if "cats" in model:
  for col in model["cats"]:
   cat = model["cats"][col]
   lines.append(col+"___RELA = "+radify_cat(cat, col)+";")
   combLine+=(fc+col+"___RELA")
 
 if "conts" in model:
  for col in model["conts"]:
   cont = model["conts"][col]
   lines.append(col+"___RELA = "+radify_cont(cont, col, defaultValue)+";")
   combLine+=(fc+col+"___RELA")
 
 if "catcats" in model:
  for cols in model["catcats"]:
   col1, col2 = cols.split(" X ")
   lines.append(col1+"__X__"+col2+"___RELA = "+radify_catcat(model, cols)+";")
   combLine+=(fc+col1+"__X__"+col2)
 
 if "catconts" in model:
  for cols in model["catconts"]:
   col1, col2 = cols.split(" X ")
   lines.append(col1+"__X__"+col2+"___RELA = "+radify_catcont(model, cols, defaultValue)+";")
   combLine+=(fc+col1+"__X__"+col2)
 
 if "contconts" in model:
  print("WARNING: radification of cont-cont interactions is untested at time of writing; caveat!")
  for cols in model["contconts"]:
   col1, col2 = cols.split(" X ")
   lines.append(col1+"__X__"+col2+"___RELA = "+radify_contcont(model, cols, defaultValue)+";")
   combLine+=(fc+col1+"__X__"+col2)
 
 
 combLine+=";"
 lines.append(combLine)
 
 
 if logistic:
  lines.append("PREDICTION = 1/(1+EXP(-PREDICTION));")
 else:
  lines.append("PREDICTION = COMB;")
 
 #Write to file
 
 f = open(filename, "w")
 for line in lines:
  print(line)
  f.write(line+'\n')
 f.close()

if __name__ == '__main__':
 exampleModel = {"BASE_VALUE":1700,"conts":{"cont1":[[0.01, 1],[0.02,1.1], [0.03, 1.06]], "cont2":[[37,1.2],[98, 0.9]], "cont3":[[12,1.2],[13, 0.9],[14, 1.1]]}, "cats":{"cat1":{"uniques":{"wstfgl":1.05, "florpalorp":0.92}, "OTHER":1.01},"cat2":{"uniques":{"ruska":1.1, "roma":0.93, 'rita':0.3}, "OTHER":1.9}}, 'catcats':{'cat1 X cat2':{"uniques":{"wstfgl":{"uniques":{"ruska":1.1, "roma":0.93, 'rita':0.3}, "OTHER":1.9}, "florpalorp":{"uniques":{"ruska":1.1, "roma":0.93, 'rita':0.3}, "OTHER":1.9}}, "OTHER":{"uniques":{"ruska":1.1, "roma":0.93, 'rita':0.3}, "OTHER":1.9}}}, 'catconts':{"cat1 X cont1":{"uniques":{"wstfgl":[[0.01, 1],[0.02,1.1], [0.03, 1.06]], "florpalorp":[[0.01, 1],[0.02,1.1], [0.03, 1.06]]}, "OTHER":[[0.01, 1],[0.02,1.1], [0.03, 1.06]]}}, 'contconts':{'cont1 X cont2':[[0.01, [[37,1.2],[98, 0.9]]],[0.02,[[37,1.2],[98, 0.9]]], [0.03, [[37,1.2],[98, 9000]]]]}}
 
 radify_model(exampleModel, "rm.txt")



import pandas as pd
import numpy as np

import copy

#Drag to default

def move_to_default(x, l, defaultValue=1):
 if x>defaultValue:
  return max(x-l, defaultValue)
 if x<defaultValue:
  return min(x+l, defaultValue)
 return x

def get_penalize_nondefault(pen, defaultValue=1, specificPenas={}):
 
 def penalize_nondefault(model):
  
  print("OMG HAIII!")
  if "cats" in model:
   for cat in model["cats"]:
   
    if cat in specificPenas:
     l = specificPenas[cat]
    else:
     l = pen
    
    for u in model["cats"][cat]["uniques"]:
     model["cats"][cat]["uniques"][u] = move_to_default(model["cats"][cat]["uniques"][u], l, defaultValue)
    model["cats"][cat]["OTHER"] = move_to_default(model["cats"][cat]["OTHER"], l, defaultValue)
  
  if "conts" in model:
   for cont in model["conts"]:
    
    if cont in specificPenas:
     l = specificPenas[cont]
    else:
     l = pen
    
    for i in range(len(model["conts"][cont])):
     model["conts"][cont][i][1] = move_to_default(model["conts"][cont][i][1], l, defaultValue)
  
  if "catcats" in model:
   for catcat in model["catcats"]:
    
    if catcat in specificPenas:
     l = specificPenas[catcat]
    else:
     l = pen
    
    for u1 in model["catcats"][catcat]["uniques"]:
     model["catcats"][catcat]["uniques"][u1]["OTHER"] = move_to_default(model["catcats"][catcat]["uniques"][u1]["OTHER"], l, defaultValue)
     for u2 in model["catcats"][catcat]["uniques"][u1]["uniques"]:
      model["catcats"][catcat]["uniques"][u1]["uniques"][u2] = move_to_default(model["catcats"][catcat]["uniques"][u1]["uniques"][u2], l, defaultValue)
    for u2 in model["catcats"][catcat]["OTHER"]["uniques"]:
     model["catcats"][catcat]["OTHER"]["uniques"][u2] = move_to_default(model["catcats"][catcat]["OTHER"]["uniques"][u2], l, defaultValue)
    model["catcats"][catcat]["OTHER"]["OTHER"]=move_to_default(model["catcats"][catcat]["OTHER"]["OTHER"], l, defaultValue)
  
  if "catconts" in model:
   for catcont in model["catconts"]:
    
    if catcont in specificPenas:
     l = specificPenas[catcont]
    else:
     l = pen
    
    for u in model["catconts"][catcont]["uniques"]:
     for i in range(len(model["catconts"][catcont]["uniques"][u])):
      model["catconts"][catcont]["uniques"][u][i][1] = move_to_default(model["catconts"][catcont]["uniques"][u][i][1], l, defaultValue)
    for i in range(len(model["catconts"][catcont]["OTHER"])):
     model["catconts"][catcont]["OTHER"][i][1] = move_to_default(model["catconts"][catcont]["OTHER"][i][1], l, defaultValue)
  
  if "contconts" in model:
   for contcont in model["contconts"]:
    
    if contcont in specificPenas:
     l = specificPenas[contcont]
    else:
     l = pen
    
    for i in range(len(model["contconts"][contcont])):
     for j in range(len(model["contconts"][contcont][i][1])):
      model["contconts"][contcont][i][1][j][1] = move_to_default(model["contconts"][contcont][i][1][j][1], l, defaultValue)
  
  return model
 
 return penalize_nondefault

#Combat continuous complexity

def move_to_target(x, l, t):
 if (x<t):
  return min(x+l, x+(t-x)/2)
 if (x>t):
  return max(x-l, x-(x-t)/2)
 return x

def penalize_cont_complexity(cont, pen):
 if (len(cont)>2):
  targets = []
  targets.append(cont[1][1] - (cont[2][1]-cont[1][1])/(cont[2][0]-cont[1][0])*(cont[1][0]-cont[0][0]))
  for i in range(len(cont)-2):
   targets.append(cont[i][1] + (cont[i+2][1]-cont[i][1])/(cont[i+2][0]-cont[i][0])*(cont[i+1][0]-cont[i][0]))
  targets.append(cont[-2][1] + (cont[-2][1]-cont[-3][1])/(cont[-2][0]-cont[-3][0])*(cont[-1][0]-cont[-2][0]))
  
  for i in range(len(cont)):
   cont[i][1] = move_to_target(cont[i][1], pen, targets[i])
  
  return cont

def get_penalize_conts_complexity(pen):
 def penalize_conts_complexity(model):
  for c in model["conts"]:
   model["conts"][c] = penalize_cont_complexity(model["conts"][c], pen)
  return model
 return penalize_conts_complexity

#Looping conts

def get_enforce_loops(cols):
 
 def enforce_loops(model):
  opModel = copy.deepcopy(model)
  for c in cols:
   opModel["conts"][c][0][1] = (model["conts"][c][0][1] + model["conts"][c][-1][1])/2
   opModel["conts"][c][-1][1] = (model["conts"][c][0][1] + model["conts"][c][-1][1])/2
  return opModel
 
 return enforce_loops

#Minimum relativities; useful for preventing multipliers from going <0

def get_enforce_min_rela(minRela=0.1):
 
 def enforce_min_rela(model): #Not expanded to interxes 
  
  opModel = copy.deepcopy(model)
  
  if "conts" in opModel:
   for col in opModel["conts"]:
    for i in range(len(opModel["conts"][col])):
     opModel["conts"][col][i][1] = max(minRela, opModel["conts"][col][i][1])
  
  if "cats" in opModel:
   for col in opModel["cats"]:
    for u in opModel["cats"][col]["uniques"]:
     opModel["cats"][col]["uniques"][u] = max(minRela, opModel["cats"][col]["uniques"][u])
    opModel["cats"][col]["OTHER"] = max(minRela, opModel["cats"][col]["OTHER"])
  
  return opModel
 
 return enforce_min_rela


#Monotony

def gently_monotonize_this_list(l, increasing=True, r=10):
 
 newL = l.copy()
 
 curseList = [False]*len(l)
 for i in range(len(newL)-1):
  if ((l[i+1]<l[i]) and increasing) or ((l[i+1]>l[i]) and not increasing):
   curseList[i]=True
   curseList[i+1]=True
 
 i = 0
 while i < len(l):
  if curseList[i]:
   start = i
   while (i < len(newL)) and (curseList[i]):
    i += 1
   end = i
   avg = sum(newL[start:end]) / (end - start)
   for j in range(start, end):
    newL[j] = avg
  else:
   i += 1
 
 if r<=1:
  return newL
 else:
  return gently_monotonize_this_list(newL, increasing, r-1)

def firmly_monotonize_this_list(l, increasing=True):
 newL = l.copy()
 if increasing:
  for i in range(len(newL)-1):
   if newL[i+1]<newL[i]:
    newL[i+1]=newL[i]
 else:
  for i in range(len(newL)-1):
   if newL[i+1]>newL[i]:
    newL[i+1]=newL[i]
 return newL

def monotonize_this_list(l, increasing=True, r=10):
 return firmly_monotonize_this_list(gently_monotonize_this_list(l, increasing, r),increasing)

def monotonize_this_cat(model, col, increasing, r=10):
 newModel = copy.deepcopy(model)
 
 k = list(model["cats"][col]["uniques"].keys())
 v = list(model["cats"][col]["uniques"].values())
 newv = monotonize_this_list(v, increasing, r)
 
 for i in range(len(newv)):
  newModel["cats"][col]["uniques"][k[i]]=newv[i]
 
 return newModel

def monotonize_this_cont(model, col, increasing, r=10):
 newModel = copy.deepcopy(model)
 
 v = [x[1] for x in model["conts"][col]]
 newv = monotonize_this_list(v, increasing, r)
 
 for i in range(len(newv)):
  newModel["conts"][col][i][1]=newv[i]
 
 return newModel

def get_monotonize_this_model(catIncDict, contIncDict, r=10):
 def monotonize_this_model(model):
  newModel = copy.deepcopy(model)
  for col in catIncDict:
   newModel = monotonize_this_cat(newModel, col, catIncDict[col], r)
  for col in contIncDict:
   newModel = monotonize_this_cont(newModel, col, contIncDict[col], r)
  return newModel
 return monotonize_this_model




import pandas as pd
import numpy as np
import math
import scipy
from scipy.special import erf

#Easy objective functions

def Gauss_grad(pred,act):
 return 2*(pred-act)

def Poisson_grad(pred,act):
 return (pred-act)/pred

def Gamma_grad(pred,act):
 return (pred-act)/(pred*pred)

def Logistic_grad(pred,act):
 return (pred-act)/(pred*(1-pred))

#Tweedie

def produce_Tweedie_grad(pTweedie=1.5):
 def Tweedie_grad(pred,act):
  return (pred-act)/(pred**pTweedie)
 return Tweedie_grad

#Easy linkages


def Unity_link(x):
 return x

def Unity_link_grad(x):
 return 1

def Unity_delink(x):
 return x


def Root_link(x):
 return x*x

def Root_link_grad(x):
 return 2*x

def Root_delink(x):
 return x**0.5


def Log_link(x):
 return np.exp(x)

def Log_link_grad(x):
 return np.exp(x)

def Log_delink(x):
 return np.log(x)


def Logit_link(x):
 return 1/(1+np.exp(-x))

def Logit_link_grad(x):
 return np.exp(-x)/((1+np.exp(-x))**2)

def Logit_delink(x):
 return np.log(x/(1-x))

links={"Unity":Unity_link,"Root":Root_link,"Log":Log_link,"Logit":Logit_link}
linkgrads={"Unity":Unity_link_grad,"Root":Root_link_grad,"Log":Log_link_grad,"Logit":Logit_link_grad}
delinks = {"Unity":Unity_delink, "Root":Root_delink, "Log":Log_delink, "Logit":Logit_delink}

#Multiple modelling!

def Take0(*args):
 return args[0]
def Take1(*args):
 return args[1]
def Take2(*args):
 return args[2]
def Take3(*args):
 return args[3]
def Take4(*args):
 return args[4]

def js0(*args):
 return pd.Series([0]*len(args[0]))
def js1(*args):
 return pd.Series([1]*len(args[0]))

def nonefunc(*args):
 return None


def Add_mlink(*args):
 return sum(args) #Amazingly, this works!

def Add_mlink_grad(*args):
 return pd.Series([1]*len(args[0]))


def Max_mlink_2(x1, x2):
 return (x1>x2)*x1+(x1<=x2)*x2 #I am aware that this line in particular is an abomination, and refuse to care. Rewrite it at your whim, dear reader.

def Max_mlink_grad_2_A(x1,x2):
 return (x1>=x2).astype(int)

def Max_mlink_grad_2_B(x1,x2):
 return (x2>=x1).astype(int)


def Min_mlink_2(x1, x2):
 return (x1<x2)*x1+(x1>=x2)*x2 #I am aware that this line in particular is an abomination, and refuse to care. Rewrite it at your whim, dear reader.

def Min_mlink_grad_2_A(x1,x2):
 return (x1<=x2).astype(int)

def Min_mlink_grad_2_B(x1,x2):
 return (x2<=x1).astype(int)


def Mult_mlink_2(x1, x2):
 return x1*x2

def Mult_mlink_grad_2_A(x1,x2):
 return x2

def Mult_mlink_grad_2_B(x1,x2):
 return x1


#Learning Rate Adjustments

def default_LRA(*args):
 return 1


def addsmoothing_LRA_0(*args):
 return sum([sum(arg.abs()) for arg in args])/sum(args[0].abs())

def addsmoothing_LRA_1(*args):
 return sum([sum(arg.abs()) for arg in args])/sum(args[1].abs())

def addsmoothing_LRA_2(*args):
 return sum([sum(arg.abs()) for arg in args])/sum(args[2].abs())

def addsmoothing_LRA_3(*args):
 return sum([sum(arg.abs()) for arg in args])/sum(args[3].abs())

def addsmoothing_LRA_4(*args):
 return sum([sum(arg.abs()) for arg in args])/sum(args[4].abs())

addsmoothing_LRAs = [addsmoothing_LRA_0,addsmoothing_LRA_1,addsmoothing_LRA_2,addsmoothing_LRA_3,addsmoothing_LRA_4]

#alongside-error models . . .

def Add_mlink_allbutlast(*args):
 return sum(args[:-1])

def Add_mlink_onlylast(*args):
 return args[-1]

def Add_mlink_grad_void(*args):
 return pd.Series([0]*len(args[0]))



def addsmoothing_LRA_0_erry(*args):
 return sum([sum(arg.abs()) for arg in args[:-1]])/sum(args[0].abs())

def addsmoothing_LRA_1_erry(*args):
 return sum([sum(arg.abs()) for arg in args[:-1]])/sum(args[1].abs())

def addsmoothing_LRA_2_erry(*args):
 return sum([sum(arg.abs()) for arg in args[:-1]])/sum(args[2].abs())

def addsmoothing_LRA_3_erry(*args):
 return sum([sum(arg.abs()) for arg in args[:-1]])/sum(args[3].abs())

def addsmoothing_LRA_4_erry(*args):
 return sum([sum(arg.abs()) for arg in args[:-1]])/sum(args[4].abs())

addsmoothing_LRAs_erry = [addsmoothing_LRA_0_erry, addsmoothing_LRA_1_erry, addsmoothing_LRA_2_erry, addsmoothing_LRA_3_erry, addsmoothing_LRA_4_erry]


#Cratio

def Cratio_mlink(*args):
 return sum(args)/(sum(args)+1)

def Cratio_mlink_grad(*args):
 return 1/((sum(args)+1)*(sum(args)+1))


#Tobit



def gnormal_u_diff(u, p, y):
 return -(y*(y-u)/((p**2)*(u**3)) - 1/u)

def gnormal_p_diff(u, p, y):
 return -((y-u)**2/((p**3)*(u**2)) - 1/p)

def PDF(u, p, y):
 return np.exp(-0.5*((y-u)/(p*u))**2) / (p*u*math.sqrt(2*math.pi))

def CDF(u, p, y):
 return 0.5*(1 - erf((y-u)/(p*u*math.sqrt(2))))

def gnormal_u_diff_censored(u, p, y):
 return -((y/u)*PDF(u,p,y)/CDF(u,p,y))

def gnormal_p_diff_censored(u, p, y):
 return -(((y-u)/p)*PDF(u,p,y)/CDF(u,p,y))

# . . . and make it double!

def gnormal_u_diff_doublecensored(u,p,y,c1,c2):
 return ((-c2/u)*PDF(u,p,c2) - (-c1/u)*PDF(u,p,c1)) / (CDF(u,p,c2) - CDF(u,p,c1))

def gnormal_p_diff_doublecensored(u,p,y,c1,c2):
 return ((-(c2-u)/p)*PDF(u,p,c2) - (-(c1-u)/p)*PDF(u,p,c1)) / (CDF(u,p,c2) - CDF(u,p,c1))

def gnormal_u_diff_doubleuncensored(u,p,y,c1,c2):
 return gnormal_u_diff(u,p,y)

def gnormal_p_diff_doubleuncensored(u,p,y,c1,c2):
 return gnormal_p_diff(u,p,y)

#Suvec

def suvec_normal_upper(pred,act):
 return (pred-act)*(pred>act)

def suvec_normal_lower(pred,act):
 return (pred-act)*(pred<act)

def suvec_gamma_upper(pred,act):
 return ((pred-act)/(pred*pred))*(pred>act)

def suvec_gamma_lower(pred,act):
 return ((pred-act)/(pred*pred))*(pred<act)

#Techne

def get_effect_of_this_cont_col(ser, cont):
 x = ser
 effectOfCol = pd.Series([1]*len(ser))
 effectOfCol.loc[(x<=cont[0][0])] = cont[0][1] #Everything too early gets with the program
 for i in range(len(cont)-1):
  x1 = cont[i][0]
  x2 = cont[i+1][0]
  y1 = cont[i][1]
  y2 = cont[i+1][1]
  effectOfCol.loc[(x>=x1)&(x<=x2)] = ((x-x1)*y2 + (x2-x)*y1)/(x2 - x1)
 effectOfCol.loc[x>=cont[-1][0]] = cont[-1][1] #Everything too late gets with the program
 return effectOfCol

def techne_mult_eval(pred,act, obj=[[-1,-1],[0,1],[1,-1]]):
 err = (pred-act)/act
 return sum(get_effect_of_this_cont_col(err, obj))/len(err)

def create_techne_mult_grad(obj):
 def techne_grad(pred, act):
  err = (pred-act)/act
  op = pd.Series([0]*len(err))
  op[err<=obj[0][0]] = (obj[1][1]-obj[0][1])/(obj[1][0]-obj[0][0])
  for i in range(len(obj)-1):
   subset = (err>=obj[i][0]) & (err<=obj[i+1][0])
   op[subset] = (obj[i+1][1]-obj[i][1])/(obj[i+1][0]-obj[i][0])
  op[err>=obj[-1][0]] = (obj[-1][1]-obj[-2][1])/(obj[-1][0]-obj[-2][0])
  return op
 return techne_grad

#Dodel Template (realistically you'll have to customize this for each model you're optimizing a dodel for)

#profit = (pri-lia)*prob
#d(profit)/d(price) = prob + (pri-lia) * d(prob)/d(price)
#d(prob)/d(price) = d(prob)/d(comb) * d(comb)/d(pricefeateff) * d(pricefeateff)/d(pricefeat) * d(pricefeat)/d(price)
#d(prob)/d(comb) is link of the comb
#d(comb)/d(pricefeateff) is, for an additive-combed model, just 1
#d(pricefeateff)/d(feat) is the model's line gradient
#d(feat)/d(price) is 1/mar because we assume a percentage

def get_dodel_lossgrad(model, df, pfname):
 def dodel_lossgrad(pri, lia, mar):
  df[pfname] = pri/mar
  
  #prob = misc.predict(model,df,"Logit")
  comb = misc.predict(model,df)
  prob = calculus.Logit_link(comb)
  
  dpdc = calculus.Logit_link_grad(comb)
  dcde = 1
  dedf = get_gradient_of_this_cont_col(pri/mar, model["conts"][pfname])
  dfdp = 1/mar
  
  dpdp = dpdc*dcde*dedf*dfdp
  
  return prob + (pri-lia) * dpdp
 return dodel_lossgrad

def get_gradient_of_this_cont_col(ser, cont):
 x = ser
 gradOfCol = pd.Series([0.0]*len(ser)) # at the points themselves, and everywhere not covered, grad is 0
 for i in range(len(cont)-1):
  x1 = cont[i][0]
  x2 = cont[i+1][0]
  y1 = cont[i][1]
  y2 = cont[i+1][1]
  gradOfCol.loc[(x>x1)&(x<x2)] = (y2 - y1)/(x2 - x1)
 return gradOfCol

#Gradient Zero Gimmick

def get_gzg_grad(obj, lb=None, ub=None):
 def gzg_grad(pred, act):
  grad = obj(pred,act)
  if ub is not None:
   grad[pred>ub]=0
  if lb is not None:
   grad[pred<lb]=0
  return grad
 return gzg_grad

